"""Load methods."""

from pathlib import Path

from termcolor import cprint

from treestamps.tree.config import TreestampsConfig
from treestamps.tree.get import TreestampsGet


class TreestampLoad(TreestampsGet):
    """Load methods."""

    def _is_path_skipped(self, path: Path) -> bool:
        """Return if path is ignored or not allowed because symlink."""
        return any(path.match(ignore_glob) for ignore_glob in self._config.ignore) or (
            not self._config.symlinks and path.is_symlink()
        )

    @classmethod
    def _load_timestamps_file_pop_and_compare_config(cls, yaml_config, compare_config):
        normalized_config = TreestampsConfig.normalize_config(yaml_config)
        # Shallow equality!
        return compare_config == normalized_config

    def _load_timestamps_file_pop_config_matches(self, yaml):
        """Return if the configured and loaded configs match."""
        yaml_ts_config = yaml.pop(self._TREESTAMPS_CONFIG_TAG, {})
        yaml_program_config = yaml.pop(self._CONFIG_TAG, None)
        return not self._config.check_config or (
            self._load_timestamps_file_pop_and_compare_config(
                yaml_ts_config, self._config.get_config_dict()
            )
            and self._load_timestamps_file_pop_and_compare_config(
                yaml_program_config, self._config.program_config
            )
        )

    def _load_timestamp_entry(
        self, timestamps_root: Path, path_str: str, ts: float
    ) -> None:
        """Load a single timestamp entry into the cache."""
        try:
            if abs_path := self._get_absolute_path(timestamps_root, path_str):
                old_ts = self.get(abs_path)
                if old_ts is None or ts > old_ts:
                    self._timestamps[abs_path] = ts
        except Exception as exc:
            cprint(f"WARNING: Invalid timestamp for {path_str}: {ts} {exc}", "yellow")

    def _load_timestamps_file(self, timestamps_path: Path) -> None:
        """Load timestamps from a file."""
        if not timestamps_path.is_file():
            return

        try:
            yaml = self._YAML.load(timestamps_path)
            if not yaml:
                return

            # Pop off config entries and compare configs.
            if not self._load_timestamps_file_pop_config_matches(yaml):
                return
            # Pop off the WAL
            wal = yaml.pop(self._WAL_TAG, ())

            # What's left are timestamp entries
            entries = yaml

            # Update entries with WAL entries
            for wal_entry in wal:
                try:
                    entries.update(wal_entry)
                except Exception as exc:
                    cprint(f"WARNING: loading WAL entry: {wal_entry}: {exc}", "yellow")

            for path_str, ts in entries.items():
                self._load_timestamp_entry(timestamps_path.parent, path_str, ts)
            if self._config.verbose:
                cprint(f"Read timestamps from {timestamps_path}")
        except Exception as exc:
            cprint(f"ERROR: parsing timestamps file: {timestamps_path} {exc}", "red")

    def _consume_child_timestamps(self, path: Path) -> None:
        """Consume a child timestamp and add its values to our root."""
        try:
            self._load_timestamps_file(path)
            if path != self._dump_path:
                self._consumed_paths.add(path)
        except Exception as exc:
            cprint(f"WARNING: reading child timestamps {exc}", "yellow")

    def _consume_all_child_timestamps(self, path: Path) -> None:
        """Recursively consume all timestamps and wal files."""
        try:
            if not path.is_dir() or self._is_path_skipped(path):
                return
            for name in (self._filename, self._wal_filename):
                self._consume_child_timestamps(path / name)
            for dir_entry in path.iterdir():
                self._consume_all_child_timestamps(dir_entry)
        except Exception as exc:
            cprint(f"WARNING: reading all child timestamps {exc}", "yellow")

    def _load_parent_timestamps(self, path: Path) -> None:
        """Recursively load timestamps from all parents."""
        if path.parent == path.parent.parent or self._is_path_skipped(path):
            return
        parent = path.parent
        self._load_timestamps_file(parent / self._filename)
        self._load_timestamps_file(parent / self._wal_filename)
        self._load_parent_timestamps(parent)

    def load(self) -> None:
        """Load all timestamps."""
        self._load_parent_timestamps(self.root_dir)
        self._consume_all_child_timestamps(self.root_dir)



"""Dump Methods."""

from collections.abc import MutableMapping
from contextlib import suppress
from pathlib import Path

from ruamel.yaml import StringIO
from termcolor import cprint

from treestamps.tree.init import TreestampsInit


class TreestampsDump(TreestampsInit):
    """Dump Methods."""

    def _get_relative_path_str(self, abs_path: Path) -> str:
        """Get the path string relative to the root_dir."""
        return str(abs_path.relative_to(self.root_dir))

    def _serialize_timestamps(self):
        """Dumpable timestamp paths need to be strings."""
        dumpable_timestamps = {}
        for abs_path, timestamp in self._timestamps.items():
            try:
                rel_path_str = self._get_relative_path_str(abs_path)
                dumpable_timestamps[rel_path_str] = timestamp
            except Exception as exc:
                cprint(f"WARNING: Serializing {abs_path}: {exc}", "yellow")
        return dumpable_timestamps

    def _get_dumpable_program_config(self) -> dict:
        """Set the config tag in the yaml to be dumped."""
        # NOTE: Treestamps symlinks & ignore options should be represented in
        # the program config.
        yaml = {}
        if self._config.program_config is not None:
            yaml[self._CONFIG_TAG] = dict(self._config.program_config)
        yaml[self._TREESTAMPS_CONFIG_TAG] = {
            "ignore": self._config.ignore,
            "symlinks": self._config.symlinks,
        }
        return yaml

    def _close_wal(self) -> None:
        """Close the write ahead log."""
        if self._wal is None:
            return
        with suppress(AttributeError):
            self._wal.close()
        self._wal = None

    def _prepare_dump(self, yaml: MutableMapping):
        config_yaml = self._get_dumpable_program_config()
        yaml.update(config_yaml)
        self._close_wal()

    def _dumpf(self, yaml: MutableMapping, path: Path):
        """Dump to file."""
        self._prepare_dump(yaml)
        self._YAML.dump(yaml, path)

    def cleanup_old_timestamps(self):
        """Cleanup old timestamps from the disk."""
        if not self._consumed_paths:
            return
        self._consumed_paths.discard(self._dump_path)
        for path in self._consumed_paths:
            try:
                path.unlink(missing_ok=True)
            except Exception as exc:
                cprint(f"WARNING: removing old timestamp: {exc}", "yellow")
        self._consumed_paths = set()

    def dumpf(self) -> None:
        """Serialize timestamps and dump to file."""
        dumpable_timestamps = self._serialize_timestamps()
        self._dumpf(dumpable_timestamps, self._dump_path)
        self.cleanup_old_timestamps()

    def dump(self) -> None:
        """Compatibility alias for dumpf()."""
        self.dumpf()

    def dumps(self):
        """Dump to string."""
        yaml = self._serialize_timestamps()
        self._prepare_dump(yaml)
        with StringIO() as buf:
            self._YAML.dump(yaml, buf)
            return buf.getvalue()



"""Set Methods."""

from datetime import datetime, timezone
from pathlib import Path

from termcolor import cprint

from treestamps.tree.dump import TreestampsDump


class TreestampsSet(TreestampsDump):
    """Set Methods."""

    _WAL_HEADER = TreestampsDump._WAL_TAG + ":\n"  # noqa: SLF001

    def _compact_timestamps_below(self, abs_root_path: Path) -> None:
        """Compact the timestamp cache below a particular path."""
        if not abs_root_path.is_dir():
            return
        root_timestamp = self._timestamps.get(abs_root_path)
        if root_timestamp is None:
            return
        delete_paths = set()
        for abs_path, timestamp in self._timestamps.items():
            if (
                abs_path.is_relative_to(abs_root_path) and timestamp < root_timestamp
            ) or timestamp is None:
                delete_paths.add(abs_path)
        for del_path in delete_paths:
            del self._timestamps[del_path]
        if self._config.verbose > 1:
            cprint(f"Compacted timestamps under: {abs_root_path}: {root_timestamp}")

    def _write_ahead_log(self, abs_path, mtime):
        """Write to the WAL."""
        if not self._wal:
            # init wall
            self._dumpf({}, self._wal_path)
            self._consumed_paths.add(self._wal_path)
            self._wal = self._wal_path.open("a")
            self._wal.write(self._WAL_HEADER)

        # Manually construct yaml dict list item.
        path_str = self._get_relative_path_str(abs_path)
        if ":" in path_str:
            # Safe wal strings. Handled automatically by yaml dumper.
            # https://github.com/commx/ruamel-yaml/blob/master/util.py#L211
            path_str = "'" + path_str.replace("'", "''") + "'"
        wal_entry = f"- {path_str}: {mtime}\n"

        self._wal.write(wal_entry)

    def set(
        self,
        path: Path | str,
        mtime: float | None = None,
        compact: bool = False,  # noqa: FBT002
    ) -> float | None:
        """Record the timestamp."""
        abs_path = self._get_absolute_path(self.root_dir, path)
        if not abs_path:
            return None

        # Should we do the set?
        old_mtime = self._timestamps.get(abs_path)
        if mtime is None:
            mtime = datetime.now(tz=timezone.utc).timestamp()
        if old_mtime and old_mtime > mtime:
            return None

        # Set timestamp
        self._timestamps[abs_path] = mtime

        # compact
        if compact:
            self._compact_timestamps_below(abs_path)

        # write to wal
        self._write_ahead_log(abs_path, mtime)
        return mtime



"""A dict of Treestamps."""

from collections.abc import Iterable
from copy import copy
from dataclasses import asdict, dataclass
from pathlib import Path

from termcolor import cprint

from treestamps.config import CommonConfig
from treestamps.tree import Treestamps
from treestamps.tree.init import TreestampsConfig


@dataclass
class GrovestampsConfig(CommonConfig):
    """Grovestamps config."""

    paths: Iterable[str | Path] = ()

    def __post_init__(self):
        """
        Pathify, filter, dedupe, order and tuplify paths.

        This order creates dir based treestamps before files so dirs get children
        recursed and files only don't.
        """
        super().__post_init__()
        dirs: set[Path] = set()
        files: set[Path] = set()
        for path_str in self.paths:
            path = Path(path_str)
            if not self.symlinks and path.is_symlink():
                continue
            if path.is_dir():
                dirs.add(path)
            else:
                files.add(path)
        self.paths = tuple(sorted(dirs) + sorted(files))

    def get_treestamps_config_dict(self):
        """Get a treestamps style config dict from this config."""
        config = copy(self)
        if config.program_config is not None:
            config.program_config = dict(config.program_config)
        config_dict = asdict(config)
        config_dict.pop("paths", None)
        return config_dict


class Grovestamps(dict):
    """A path keyed dict of Treestamps."""

    def __init__(self, config: GrovestampsConfig) -> None:
        """Create a dictionary of Treestamps keyed with paths."""
        self._config = config

        treestamps_config_dict = self._config.get_treestamps_config_dict()

        for top_path in self._config.paths:
            root_dir = Treestamps.get_dir(top_path)
            if root_dir in self:
                continue
            tree_config = TreestampsConfig(
                **treestamps_config_dict, path=Path(top_path)
            )
            ts = Treestamps(tree_config)
            ts.load()
            self[root_dir] = ts

    def dump(self) -> None:
        """Dump all treestamps."""
        for top_path, treestamps in self.items():
            if self._config.verbose:
                cprint(f"Saving timestamps for {top_path}")
            treestamps.dump()
